from igraph import *
from enum import Enum
import itertools
import sys


# input is graph and evaluate returns the "value" of this graph
# problem is integer coding of different problems
#  problem 1 is densest k subgraph - it return number of edges
#  problem 2 is maximizing of min-degree in the subgraph

class Fco(Enum):
    DENSEST = 1  # densest k subgraph
    MINDEG = 2  # maximize mindegree in subgraph
    MINMAXDEG = 3  # minimize maxdegree in subgraph
    ACYCLIC = 4  # find acyclic subgraph
    TRIAFREE = 5  # find triangle free subgraph
    MAXDIAM = 6  # maximize diameter subgraph
    RREG = 7  # find r-regular subgraph for parameter r
    BOUNDDEG = 8  # find subgraph such that each degree d fulfills: c_1 <= d <= c_2 for parameters c_1 and c_2

# problem
problem = Fco.DENSEST
# parameter-vector for the problem (may be empty)
parameters = []
# adding an edge changes value by 0 or a positive number
edge_monotone = False
# adding a vertex can increase the value by up to c, if this value is equal to -1, this property is inactive
vertex_upper_bound = -1
# parameter k
k = 1
# v

# user provided upper bound
user_ub = sys.maxsize


def check_user_ub(value):
    # type: (object) -> object
    global user_ub
    if value >= user_ub:
        return True
    return False


def set_problem(prob, paras, k_in):
    global problem
    global parameters
    global edge_monotone
    global vertex_upper_bound
    global user_ub
    global k
    problem = Fco(prob)
    parameters = list(paras)
    k = k_in
    # densest k-subgraph
    if problem == Fco.DENSEST:
        edge_monotone = True
        vertex_upper_bound = k
        user_ub = k*(k-1)/2
    # maximize min-degree
    elif problem == Fco.MINDEG:
        edge_monotone = True
        vertex_upper_bound = 1
        user_ub = k-1
    # minimize max-degree
    elif problem == Fco.MINMAXDEG:
        edge_monotone = False
        vertex_upper_bound = 0
        user_ub = -2
    # find an acyclic graph
    elif problem == Fco.ACYCLIC:
        edge_monotone = False
        vertex_upper_bound = 0
        user_ub = 1
    # find a triangle free graph
    elif problem == Fco.TRIAFREE:
        edge_monotone = False
        vertex_upper_bound = 0
        user_ub = 1
    # maximize the diameter
    elif problem == Fco.MAXDIAM:
        edge_monotone = False
        vertex_upper_bound = 1
        user_ub = k-1
    # find r-regular graph (aditional parameter r)
    elif problem == Fco.RREG:
        edge_monotone = False
        vertex_upper_bound = 1
        user_ub = 1
    # find graph with min-deg >= a and max-deg <= b (aditional parameters a and b, in this sequence)
    elif problem == Fco.BOUNDDEG:
        edge_monotone = False
        vertex_upper_bound = 1
        user_ub = 1


def evaluate(graph,host_graph,forbidden,list_p):
    global problem
    global parameters
    global k
    # densest k-subgraph
    if problem == Fco.DENSEST:
        return graph.ecount()
    # maximize min-degree
    elif problem == Fco.MINDEG:
        min_deg = graph.vcount()
        for i in range(graph.vcount()):
            min_deg = min(graph.degree(i),min_deg)
        return min_deg
    # minimize max-degree
    elif problem == Fco.MINMAXDEG:
        return - graph.maxdegree()
    # find an acyclic graph
    elif problem == Fco.ACYCLIC:
        if graph.ecount() == graph.vcount() - 1:
            return 1
        else:
            return 0
    # find a triangle free graph
    elif problem == Fco.TRIAFREE:
        for edge in graph.es:
            # check if neighborhoods of endpoints are not disjoint
            if set(graph.neighbors(edge.source)).isdisjoint(set(graph.neighbors(edge.target))) == 0:
                return 0
        return 1
    # maximize the diameter
    elif problem == Fco.MAXDIAM:
        return graph.diameter(directed=False)
    # find r-regular graph (aditional parameter r)
    elif problem == Fco.RREG:
        # parameter r (this is the only value in parameters)
        r = parameters[0]
        parameters.append(r)
        problem = Fco.BOUNDDEG
    # find graph with min-deg >= a and max-deg <= b (additional parameters a and b)
    if problem == Fco.BOUNDDEG:
        a=parameters[0]
        b=parameters[1]
        n=graph.vcount()
        host_ids = set()
        saturated_ids = set()
        #check: some vertex has too high degree
        if graph.maxdegree() > b:
            return -sys.maxsize
        for i in range(graph.vcount()):
            i_host = graph.vs[i]["id"]
            if i_host != None:
                #check: some vertex has too low degree and may not get further neighbors        
                if graph.degree(i) < a and i not in list_p:
                    return -sys.maxsize
                #check: some vertex has too low degree and may not get enough neighbors        
                if graph.degree(i)+(k-n) < a:
                    return -sys.maxsize
                #check: some vertex has too low degree in the host graph        
                if host_graph.degree(i_host) < a:
                    return -sys.maxsize
                else:
                    host_ids.add(i_host)
                    if graph.degree(i)==b:
                        saturated_ids.add(i_host)
        if n<k:
            demand = 0
            safe = False
            for i in range(graph.vcount()):
                # if list_p has only vertices with degree b, discard this subgraph
                if graph.degree(i)<b and i in list_p:
                    safe = True
                # sum the demand of all vertices                
                if graph.degree(i) < a:
                    demand += a-graph.degree(i)
            if not safe:
                return -sys.maxsize
            # discard subgraph if even if all further vertices will not be sufficient for the demand 
            if demand > (k-n)*b:
                return -sys.maxsize
            # now we will have a closer look at demand vertices 
            for i in range(n):
                i_host = graph.vs[i]["id"]
                if (i_host)==None or graph.degree(i) >= a:
                    continue
                # Ni will contain all neighbors of i that can be added
                Ni =  set()
                for v in host_graph.neighbors(i_host):
                   Nv = set(host_graph.neighbors(v))
                   if host_graph.degree(v) >= a and v not in host_ids and not forbidden[v] and len(Nv.intersection(host_ids))<=b:
                       Ni.add(v)
                if (a > graph.degree(i)+len(Ni)):
                       return -sys.maxsize
                for j in range(n):
                   j_host = graph.vs[j]["id"]
                   if j==i or j_host==None or graph.degree(j)<b:
                       continue
                   Ni.difference_update(host_graph.neighbors(j_host))
                   if (a > graph.degree(i)+len(Ni)):
                       return -sys.maxsize
                for j in range(i+1,n):
                   j_host = graph.vs[j]["id"]
                   if (j_host)==None or graph.degree(j)>=a:
                       continue
                   Nj = set()
                   for v in host_graph.neighbors(j_host):
                       if host_graph.degree(v) >= a and not forbidden[v]:
                           Nj.add(v)
                   Nij=Ni.intersection(Nj)
                   if (2*a- graph.degree(i)-graph.degree(j)-min(b- graph.degree(i),b- graph.degree(j),len(Nij))) > k-n:
                       return -sys.maxsize
                   elif 2*(2*a- graph.degree(i)-graph.degree(j)-len(Nij)) > k-n:
                       for l in range(j+1,n):
                           l_host = graph.vs[l]["id"]
                           if (l_host)==None or graph.degree(l)>=a:
                               continue
                           Nl = set() 
                           for v in host_graph.neighbors(l_host):
                               if host_graph.degree(v) >= a and not forbidden[v]:
                                   Nl.add(v)
                           if len(Ni.intersection(Nl))==0 and len (Nj.intersection(Nl))==0:
                               if (3*a- graph.degree(i)-graph.degree(j)-graph.degree(l)-min(b- graph.degree(i),b- graph.degree(j),len(Nij))) > k-n:
                                   return -sys.maxsize
                # greedy computation of vertices with disjoint neighborhoods
                Ndisjoint =  set(Ni)
                joint_deficit = a - graph.degree(i)
                for j in range(i+1,n):
                   j_host = graph.vs[j]["id"]
                   if (j_host)==None or graph.degree(j)>=a:
                       continue
                   Nj = set()
                   for v in host_graph.neighbors(j_host):
                       Nv = set(host_graph.neighbors(v))
                       if host_graph.degree(v) >= a and v not in host_ids and not forbidden[v] and len(Nv.intersection(host_ids))<=b and len(Nv.intersection(saturated_ids))==0:
                           Nj.add(v)
                   if len(Nj.intersection(Ndisjoint))==0: 
                       Ndisjoint.update(Nj)
                       joint_deficit += a - graph.degree(j)
                   if joint_deficit > k-n:
                       return -sys.maxsize
                   
        for i in range(graph.vcount()):
            if a > graph.degree(i):
                return 0
        return 1
    else:
        return 0

# calculate better upper bounds for subgraphs for edge-monotone properties
def eval_ub(graph, host_graph, forbidden, list_p, s, value, cb_value):
    global problem
    if problem == Fco.DENSEST:
        forb_vertices = set()
        neighborhood = set()
        # second neighborhood
        neigs_2 = set()
        host_p = set()
        for entry in list_p:
            id = graph.vs[entry]["id"]
            host_p.update({id})
            for vertex in host_graph.neighbors(id):
                if forbidden[vertex] is False:
                    neighborhood.add(vertex)
                    for neig in host_graph.neighbors(vertex):
                        if forbidden[neig] is True:
                            forb_vertices.add(neig)
                        else:
                            neigs_2.add(neig)
                else:
                    forb_vertices.add(vertex)
        values_2 = [[] for i in range(s)]
        for vertex in neigs_2:
            neigs_in_np = len(set(host_graph.neighbors(vertex)).intersection(neighborhood))
            for s_prime in range(s):
                values_2[s_prime].append(min(neigs_in_np, s - s_prime))
        values_s = [[] for i in range(s)]
        for vertex in neighborhood:
            neigs_in_p = len(host_p.intersection(host_graph.neighbors(vertex)))
            neigs_out = set(host_graph.neighbors(vertex)).difference(forb_vertices)
            neigs_in_np = neigs_out.intersection(neighborhood)
            neigs_in_np = len(neigs_in_np)
            # update values
            for s_prime in range(s):
                values_s[s_prime].append(neigs_in_p + float(min(neigs_in_np, s-s_prime-1))/2)
        # sort each of the s arrays and pick the best values
        upper_bound = 0
        for s_prime in range(s):
            vals = sorted(values_s[s_prime], reverse=True)[:(s-s_prime)]
            vals2 = sorted(values_2[s_prime], reverse=True)[:s_prime]
            upper_bound = max(upper_bound, int(sum(vals)) + int(sum(vals2)) + float(s_prime*(s_prime-1))/2)
        if upper_bound + value <= cb_value:
            return True
    elif problem == Fco.MINDEG:
        forb_vertices = set()
        neighborhood = set()
        host_p = set()
        for entry in list_p:
            id = graph.vs[entry]["id"]
            # if degree of this vertex is to small, abort
            if host_graph.degree(id) <= cb_value:
                return True
            host_p.update({id})
            for vertex in host_graph.neighbors(id):
                if forbidden[vertex] is False:
                    neighborhood.add(vertex)
                    for neig in host_graph.neighbors(vertex):
                        if forbidden[neig] is True:
                            forb_vertices.add(neig)
                else:
                    forb_vertices.add(vertex)
        # check degree without the forbidden vertices
        for entry in list_p:
            if len(set(host_graph.neighbors(graph.vs[entry]["id"])).difference(forb_vertices)) <= cb_value - graph.degree(entry):
                return True
        values = 0
        for vertex in neighborhood:
            neigs_in_p = host_p.intersection(host_graph.neighbors(vertex))
            neigs_in_p = len(neigs_in_p)
            neigs_out = len(set(host_graph.neighbors(vertex)).difference(forb_vertices))
            val_vertex = neigs_in_p + min(s-1, neigs_out)
            # this vertex only improves the current best value if cal_vertex > cb_value
            if val_vertex > cb_value:
                values += 1
        # it is also possible to add a clique of size s-1, if only one vertex of neighborhood is added
        if s-1 > cb_value:
            values = values + s - 1
        # if there are les than s entries in value, abort this subgraph
        if values < s:
            return True
    return False

# returns maximum "value" which can be achieved by adding s vertices
# graph is current subgraph
# paras_properties are the parameters for the properties like edge_monotone
# s is the number of missing vertices to get a subgraph of size k
# list_p is the list of vertices of the current subgraph which can have neighbors outside the current subgraph
# numb_childs is an estimation for the number of children, we need this number to know if the check with universal graphs is usefull
# cb_value is the current best value of a connected size_k subgraph
# forbidden is the current set of forbidden vertices
def eval_add_vertices(graph, host_graph, s, list_p, numb_childs, cb_value, forbidden):
    # each property calculates a upper bound
    # return the best (lowest upper bound)
    global k
    global problem
    global edge_monotone
    global vertex_upper_bound
    if len(list_p) == 0:
        return True
    valuex = evaluate(graph,host_graph,forbidden,list_p)
    if vertex_upper_bound > -1:
        value = valuex + s*vertex_upper_bound
        if value <= cb_value:
            return True
    # if edge-monotone, add a clique of size s to the graph and then evaluate
    numb_vertices = graph.vcount()
    if edge_monotone == True:
        graph.add_vertices(s)
        edges=[]
        for i in range(numb_vertices, numb_vertices+s):
            # each new vertex has an edge to each vertex in the origin graph
            for j in list_p:
                edges.append((i, j))
            # add a clique of s vertices
            for j in range(i+1, k):
                edges.append((i, j))
        graph.add_edges(edges)
        value = evaluate(graph,host_graph,forbidden,list_p)
        # remove the clique from the graph to reuse the input graph
        graph.delete_vertices(range(numb_vertices, k))
        if value <= cb_value:
            return True
        # compute an upper bound using the host graph
        return eval_ub(graph, host_graph, forbidden, list_p, s, valuex, cb_value)
    # universal graph rule will only applied if property is not edge-monotone
    else:
        # check with universal graphs for s<=3 if we can obtain a (better) upperbound, only the vertices in set_p can have neighbors outside the current subgraph
        # we will brake if we find a subgraph with a value better than the current best value
        # only do this calculation if upperbound is still sys.maxsize, since this test costs much time
        # break if we find a value which is higher than the current best value of a connected subgraph of order k
        # universal_graph_const is value between 0 and 1, a ratio when we apply this pruning rule
        numb_edges = graph.ecount()
        numb_vertex_pairs = len(list_p) * s
        if 0 < s <= 3 and numb_vertex_pairs < 26 and len(list_p)*s+s-1 < math.log((numb_childs+1)**s, 2):
            # we repeat nearly the same code 3 times, to avoid checks which would appear for each possibility
            if s == 1:
                graph.add_vertices(1)
                # try all possible edge combinations between the actual graph (set_p) and the appropriate universal graph
                for possibility in list(itertools.product([0, 1], repeat=numb_vertex_pairs)):
                    # translate possibility in edges, for each possibility 1 represents an edge and 0 a non-edge
                    # with div, we can determine the vertex in graph, and with mod the vertex in the universal graph
                    # to determine to vertex in graph, get the vertex name of the list in p and the corresponding ID
                    check = False
                    for i in range(numb_vertex_pairs):
                        if possibility[i] == 1:
                            check = True
                            # this edge is connected with exactly one vertex of the universal graph, set the index to True
                            graph.add_edge(list_p[i], numb_vertices)
                    if check == True:
                        value = evaluate(graph,host_graph,forbidden,list_p)
                        if value > cb_value:
                            graph.delete_vertices(range(numb_vertices, k))
                            return False
                    # afterwards remove the added edges and try the next possibility
                    graph.delete_edges(range(numb_edges, graph.ecount()))
                graph.delete_vertices(range(numb_vertices, k))
            elif s == 2:
                graph.add_vertices(2)
                # try all possible edge combinations between the actual graph (set_p) and the appropriate universal graph
                for possibility in list(itertools.product([0, 1], repeat=numb_vertex_pairs)):
                    # translate possibility in edges, for each possibility 1 represents an edge and 0 a non-edge
                    # with div, we can determine the vertex in graph, and with mod the vertex in the universal graph
                    # to determine to vertex in graph, get the vertex name of the list in p and the corresponding ID

                    # to save to check for connectivity, check which vertex from the uni_graph gets an edge
                    u_edges = []
                    test_conn = [False] * 2
                    for i in range(numb_vertex_pairs):
                        if possibility[i] == 1:
                            # this edge is connected with exactly one vertex of the universal graph, set the index to True
                            test_conn[i % 2] = True
                            u_edges.append((list_p[int(i / s)], numb_vertices + i % s))
                    graph.add_edges(u_edges)
                    # now check for each possible subgraph with s vertices if we obtain a better upper bound with this edge set between the actual graph and the new subgraph of s vertices
                    # check which new vertices have an edge; next choose the possible subgraphs
                    if test_conn[0] == True:
                        # try possibility with a path of 2 vertices
                        graph.add_edge(numb_vertices, numb_vertices + 1)
                        value = evaluate(graph,host_graph,forbidden,list_p)
                        if value > cb_value:
                            graph.delete_vertices(range(numb_vertices, k))
                            return False
                        graph.delete_edges(graph.ecount() - 1)
                        if test_conn[1] == True:
                            # try possibility with 2 isolated vertices
                            value = evaluate(graph,host_graph,forbidden,list_p)
                            if value > cb_value:
                                graph.delete_vertices(range(numb_vertices, k))
                                return False
                    else:
                        if test_conn[1] == True:
                            # try possibility with a path of 2 vertices
                            graph.add_edge(numb_vertices, numb_vertices + 1)
                            value = evaluate(graph,host_graph,forbidden,list_p)
                            if value > cb_value:
                                graph.delete_vertices(range(numb_vertices, k))
                                return False
                            graph.delete_edges(graph.ecount() - 1)
                    # afterwards remove the added edges and try the next possibility
                    graph.delete_edges(range(numb_edges, graph.ecount()))
                graph.delete_vertices(range(numb_vertices, k))
            else:
                graph.add_vertices(3)
                # try all possible edge combinations between the actual graph (set_p) and the appropriate universal graph
                for possibility in list(itertools.product([0, 1], repeat=numb_vertex_pairs)):
                    # translate possibility in edges, for each possibility 1 represents an edge and 0 a non-edge
                    # with div, we can determine the vertex in graph, and with mod the vertex in the universal graph
                    # to determine to vertex in graph, get the vertex name of the list in p and the corresponding ID

                    # to save to check for connectivity, check which vertex from the uni_graph gets an edge
                    test_conn = [False] * 3
                    u_edges=[]
                    for i in range(numb_vertex_pairs):
                        if possibility[i] == 1:
                            # this edge is connected with exactly one vertex of the universal graph, set the index to True
                            test_conn[i % 3] = True
                            u_edges.append((list_p[int(i / s)], numb_vertices + i % s))
                    graph.add_edges(u_edges)
                    # now check for each possible subgraph with 3 vertices if we obtain a better upper bound with this edge set between the actual graph and the new subgraph of s vertices
                    if test_conn[2] == True:
                        # try possibility with a path of 3 vertices
                        graph.add_edge(numb_vertices, numb_vertices + 1)
                        graph.add_edge(numb_vertices + 1, numb_vertices + 2)
                        value = evaluate(graph,host_graph,forbidden,list_p)
                        if value > cb_value:
                            graph.delete_vertices(range(numb_vertices, k))
                            return False
                        # try possibility with a clique of 3 vertices
                        graph.add_edge(numb_vertices, numb_vertices + 2)
                        value = evaluate(graph,host_graph,forbidden,list_p)
                        if value > cb_value:
                            graph.delete_vertices(range(numb_vertices, k))
                            return False
                        graph.delete_edges(range(graph.ecount() - 2, graph.ecount()))
                        if test_conn[0] == True:
                            value = evaluate(graph,host_graph,forbidden,list_p)
                            if value > cb_value:
                                graph.delete_vertices(range(numb_vertices, k))
                                return False
                            graph.delete_edges(graph.ecount() - 1)
                            if test_conn[1] == True:
                                value = evaluate(graph,host_graph,forbidden,list_p)
                                if value > cb_value:
                                    graph.delete_vertices(range(numb_vertices, k))
                                    return False
                        else:
                            graph.delete_edges(graph.ecount() - 1)
                    else:
                        if test_conn[0] == True or test_conn[1] == True:
                            # try possibility with a path of 3 vertices
                            graph.add_edge(numb_vertices, numb_vertices + 1)
                            graph.add_edge(numb_vertices + 1, numb_vertices + 2)
                            value = evaluate(graph,host_graph,forbidden,list_p)
                            if value < cb_value:
                                graph.delete_vertices(range(numb_vertices, k))
                                return False
                            # try possibility with a clique of 3 vertices
                            graph.add_edge(numb_vertices, numb_vertices + 2)
                            value = evaluate(graph,host_graph,forbidden,list_p)
                            if value < cb_value:
                                graph.delete_vertices(range(numb_vertices, k))
                                return False
                            graph.delete_edges(range(graph.ecount() - 3, graph.ecount()))

                    # afterwards remove the added edges and try the next possibility
                    graph.delete_edges(range(numb_edges, graph.ecount()))
                graph.delete_vertices(range(numb_vertices, k))
            return True
    return False
